from pathlib import Path
from typing import Any

import xarray as xr

from cmip_ref_core.datasets import FacetFilter, SourceDatasetType
from cmip_ref_core.metrics import DataRequirement, Metric, MetricExecutionDefinition, MetricResult
from cmip_ref_core.pycmec.metric import CMECMetric
from cmip_ref_core.pycmec.output import CMECOutput


def calculate_annual_cycle(input_files: list[Path]) -> xr.Dataset:
    """
    Calculate the annual mean timeseries for a dataset.

    While this function is implemented here,
    in most cases the metric calculation will be in the underlying benchmarking package.
    How the metric is calculated is up to the provider.

    Parameters
    ----------
    input_files
        List of input files to calculate the annual mean timeseries.

        This dataset may consist of multiple data files.

    Returns
    -------
    :
        The annual mean timeseries of the dataset
    """
    time_coder = xr.coders.CFDatetimeCoder(use_cftime=True)
    xr_ds = xr.open_mfdataset(input_files, combine="by_coords", chunks=None, decode_times=time_coder)

    annual_mean = xr_ds.resample(time="YS").mean()
    return annual_mean.mean(dim=["lat", "lon"], keep_attrs=True)


def format_cmec_output_bundle(dataset: xr.Dataset) -> dict[str, Any]:
    """
    Create a simple CMEC output bundle for the dataset.

    Parameters
    ----------
    dataset
        Processed dataset

    Returns
    -------
        A CMEC output bundle ready to be written to disk
    """
    # TODO: Check how timeseries data are generally serialised
    cmec_output = {
        "DIMENSIONS": {
            "model": {dataset.attrs["source_id"]: {}},
            "region": {"global": {}},
            "metric": {"tas": {}},
            "json_structure": [
                "model",
                "region",
                "metric",
            ],
        },
        # Is the schema tracked?
        "SCHEMA": {
            "name": "CMEC-REF",
            "package": "example",
            "version": "v1",
        },
        "RESULTS": {
            dataset.attrs["source_id"]: {"global": {"tas": 0}},
        },
    }

    return cmec_output


class AnnualCycle(Metric):
    """
    Calculate the annual cycle for a dataset
    """

    name = "PMP Annual Cycle"
    slug = "pmp-annual-cycle"

    data_requirements = (
        DataRequirement(
            source_type=SourceDatasetType.CMIP6,
            filters=(
                FacetFilter(facets={"frequency": "mon", "experiment_id": ("historical", "amip")}),
                # Ignore some experiments because they are not relevant
                FacetFilter(facets={"experiment_id": ("hist-*")}, keep=False),
            ),
            # Add cell areas to the groups
            # constraints=(AddCellAreas(),),
            # Run the metric on each unique combination of model, variable, experiment, and variant
            group_by=("source_id", "variable_id", "experiment_id", "variant_label", "member_id"),
        ),
    )

    def run(self, definition: MetricExecutionDefinition) -> MetricResult:
        """
        Run a metric

        Parameters
        ----------
        definition
            A description of the information needed for this execution of the metric

        Returns
        -------
        :
            The result of running the metric.
        """
        # This is where one would hook into however they want to run
        # their benchmarking packages.
        # cmec-driver, python calls, subprocess calls all would work

        input_datasets = definition.metric_dataset[SourceDatasetType.CMIP6]

        annual_mean_global_mean_timeseries = calculate_annual_cycle(input_files=input_datasets.path.to_list())

        # the format function actually returns the metric bundle
        cmec_metric = format_cmec_output_bundle(annual_mean_global_mean_timeseries)
        CMECMetric.model_validate(cmec_metric)

        # create a empty CMEC output dictionary
        cmec_output = CMECOutput.create_template()
        CMECOutput.model_validate(cmec_output)

        # the cmec_output_bundle and cmec_metric_bundle are required keywords, cannot be omitted
        return MetricResult.build_from_output_bundle(
            definition, cmec_output_bundle=cmec_output, cmec_metric_bundle=cmec_metric
        )
